@@ -51,8 +51,18 @@
 
 /* Performance tuning.
  * For short sentences, encoding takes more resources than it saves. If
- * this overhead is improved, this limit can be set lower. */
-#define SENTENCE_MIN_LENGTH_TRAILING_HASH 6
+ * this overhead is improved, this limit can be set lower.
+ *
+ * Update: For a consistent linkage deduplication, encoding (which
+ * included tracon sharing) should always be done. And now the overhead
+ * is negligible.
+ *
+ * Note: setting this to non-zero values disables some of the more
+ * subtle tracon encoding code, and thus can be used to create a
+ * baseline parse, skipping that code. This can be setin with the
+ * test_enabled("min-len-encoding") flag (see api.c)
+ */
+#define SENTENCE_MIN_LENGTH_TRAILING_HASH 0
 
 /* Pruning per null-count is costly for sentences whose parsing time
  * is relatively small. If a better pruning per null-count is implemented,
@@ -91,10 +101,12 @@ struct Parse_Options_s
 	short use_spell_guess; /* Up to this many spell-guesses per unknown word 7 */
 
 	/* Choice of the parser to use */
+#if USE_SAT_SOLVER
 	bool use_sat_solver;   /* Use the Boolean SAT based parser */
+#endif
 
 	/* Options governing the parser internals operation */
-	double disjunct_cost;  /* Max disjunct cost to allow */
+	float disjunct_cost;  /* Max disjunct cost to allow */
 	short min_null_count;  /* The minimum number of null links to allow */
 	short max_null_count;  /* The maximum number of null links to allow */
 	bool islands_ok;       /* If TRUE, then linkages with islands
@@ -108,9 +120,9 @@ struct Parse_Options_s
 	/* Options governing post-processing */
 	bool perform_pp_prune; /* Perform post-processing-based pruning TRUE */
 	size_t twopass_length; /* Min sent length for two-pass post processing */
-	Cost_Model cost_model; /* For sorting linkages after parsing. */
 
 	/* Options governing the generation of linkages. */
+	Cost_Model cost_model; /* For sorting linkages after parsing. */
 	size_t linkage_limit;  /* The maximum number of linkages processed 100 */
 	bool display_morphology;/* If true, print morpho analysis of words TRUE */
 
@@ -133,12 +145,14 @@ struct Sentence_s
 	Word  *word;                /* Array of words after tokenization */
 	String_set *   string_set;  /* Used for assorted strings */
 	Pool_desc * Match_node_pool;
-	Pool_desc * Table_connector_pool; /* Count memoizing memory pool */
+	Pool_desc * Table_tracon_pool; /* Count memoizing memory pool */
 	Pool_desc * wordvec_pool;   /* For tracon-word zero-count memoizing */
 	Pool_desc * Exp_pool;
 	Pool_desc * X_node_pool;
 	Pool_desc * Disjunct_pool;
 	Pool_desc * Connector_pool;
+	Pool_desc * Clause_pool;
+	Pool_desc * Tconnector_pool;
 
 	/* Connector encoding, packing & sharing. */
 	size_t min_len_encoding;     /* Encode from this sentence length. */
@@ -158,6 +172,7 @@ struct Sentence_s
 	/* Parse results */
 	int    num_linkages_found;  /* Total number before postprocessing.  This
 	                               is returned by the do_count() function */
+	bool   overflowed;          /* True, if counting overflowed. */
 	size_t num_linkages_alloced;/* Total number of linkages allocated.
 	                               the number post-processed might be fewer
 	                               because some are non-canonical */
